#include<cstdio>
#include<cmath>
#include<opencv/cv.h>
#include<opencv/cxcore.h>
#include<opencv/highgui.h>
#include"graph.h"
#include"graph.cpp"
#include"maxflow.cpp"

#define SMOOTH_COST 1
#define COLOR_BAND 256
#define DISPARITY_MAX 16
#define SAD_SIZE 5
#define CORRECTION 16

typedef Graph<float,float,float> GraphType;

int left_right(int col, int row1, int row2, CvSize size) {
	return size.width*size.height+col*(size.width-1)+row1;
}

int top_down(int col1, int col2, int row, CvSize size) {
	return size.width*size.height+(size.width-1)*size.height+col1*size.width+row;
}

IplImage* calculate_sad(IplImage *image) {
	unsigned char *temp;
	float *tempSAD;
	float SAD_SQUARE=(SAD_SIZE*2+1)*(SAD_SIZE*2+1);
	float num;
	int width,height;
	int i,j;

	CvSize size=cvGetSize(image);
	IplImage *SAD=cvCreateImage(size, IPL_DEPTH_32F, 1);
	cvSetZero(SAD);

	for(height=SAD_SIZE; height<size.height-SAD_SIZE; height++) {
		for(width=SAD_SIZE; width<size.width-SAD_SIZE; width++) {
			num=0;
			for(j=-SAD_SIZE; j<=SAD_SIZE; j++) {
				temp = &((unsigned char *)(image->imageData + image->widthStep*(height+j)))[width];

				for(i=-SAD_SIZE; i<=SAD_SIZE; i++) {
					num+=temp[i];
				}
			}

			tempSAD = &((float *)(SAD->imageData + SAD->widthStep*height))[width];
			tempSAD[0] = num/SAD_SQUARE;
		}
	}

	return(SAD);
}

IplImage* alpha_cut(IplImage *leftImage, IplImage *rightImage) {
	GraphType *g;
 	int	currentNode,x,y;
	unsigned char *temp_fp;
	float	*tempLeftSAD, *tempRightSAD;
	CvSize size = cvGetSize(leftImage);

	IplImage *imageLeftSAD;
	IplImage *imageRightSAD;
	imageLeftSAD=calculate_sad(leftImage);
	imageRightSAD=calculate_sad(rightImage);

	IplImage *fpImage = cvCreateImage(size, IPL_DEPTH_8U, 1);
	cvSetZero(fpImage);

	double LARGE_NUM=10000000.0;
	int success;
	int alpha;

	float weightSource;
	float weightSink;
	float weightLeft;
	float weightRight;
	float weightTop;
	float weightDown;

	int lrNodeNum;
	int tdNodeNum;

	do {
		success=0;

		for(alpha=0; alpha<DISPARITY_MAX; alpha++) {
			g = new GraphType(size.width*size.height+(size.width-1)*size.height+size.width*(size.height-1),(size.width-1)*size.height*2+size.width*(size.height-1)*2); 
			g -> add_node(size.width*size.height+(size.width-1)*size.height+size.width*(size.height-1)); 
			for(y=SAD_SIZE, currentNode=0; y<size.height-SAD_SIZE; y++) {
				tempLeftSAD  = &((float*)(imageLeftSAD->imageData + imageLeftSAD->widthStep*y))[DISPARITY_MAX];
				tempRightSAD = &((float*)(imageRightSAD->imageData + imageRightSAD->widthStep*y))[DISPARITY_MAX];
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];

				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE; x++) {
					weightSource = abs(tempLeftSAD[x]-tempRightSAD[x-temp_fp[x]]);
					weightSink   = abs(tempLeftSAD[x]-tempRightSAD[x-alpha]);
					g -> add_tweights( currentNode,weightSource, weightSink);
					currentNode++;
				}
			}

			for(y=SAD_SIZE; y<size.height-SAD_SIZE; y++) {
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];
				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE-1; x++) {

					if(temp_fp[x] == temp_fp[x+1]) {
						weightLeft = (float)(SMOOTH_COST * abs(temp_fp[x]   - temp_fp[x+1]));
						lrNodeNum = y*(size.width-DISPARITY_MAX*2)+x;
						g -> add_edge( lrNodeNum, lrNodeNum+1,weightLeft, weightLeft);
					} else {
 						weightSource = (float)(SMOOTH_COST * abs(temp_fp[x] - alpha));
						weightSink   = 0;
						g -> add_tweights( currentNode, weightSource, weightSink);
						weightLeft  = (float)(SMOOTH_COST * abs(temp_fp[x]   - alpha));
						weightRight = (float)(SMOOTH_COST * abs(temp_fp[x+1] - alpha));
						lrNodeNum = y*(size.width-DISPARITY_MAX*2)+x;
						g -> add_edge( currentNode, lrNodeNum, weightLeft, weightLeft );
						g -> add_edge( currentNode, lrNodeNum+1, weightRight, weightRight );
						currentNode++;
					}
				}
			}

			for(y=SAD_SIZE; y<size.height-SAD_SIZE-1; y++) {
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];
				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE; x++) {
					if(temp_fp[x] == temp_fp[x+size.width]) {
						weightTop = (float)(SMOOTH_COST * abs(temp_fp[x]   - alpha));
						tdNodeNum = y*(size.width-DISPARITY_MAX*2)+x;
						g -> add_edge( tdNodeNum, tdNodeNum+size.width, weightTop, weightTop);
					} else {
						weightSource = (float)(SMOOTH_COST * abs(temp_fp[x] - temp_fp[x+size.width]));
						weightSink   = 0;
						g -> add_tweights( currentNode,   weightSource, weightSink);

						weightTop = (float)(SMOOTH_COST * abs(temp_fp[x] - alpha));
						weightDown = (float)(SMOOTH_COST * abs(temp_fp[x+size.width] - alpha));
						lrNodeNum = y*(size.width-DISPARITY_MAX*2)+x;
						g -> add_edge( currentNode, tdNodeNum,    weightTop, weightTop );
						g -> add_edge( currentNode, tdNodeNum+size.width,   weightDown, weightDown );
						currentNode++;
					}
				}
			}

			double large_num = g->maxflow();

			if(large_num < LARGE_NUM) {
				success = 1;
				LARGE_NUM = large_num;

				for(y=SAD_SIZE, currentNode=0; y<size.height-SAD_SIZE; y++) {
					temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[0];

					for(x=DISPARITY_MAX;x<size.width-SAD_SIZE;x++) {
						if(g->what_segment(currentNode)==GraphType::SOURCE){
							temp_fp[x] = alpha;
						}

						currentNode++;
					}
				}
			}

			delete g; 
		}
	}
	while(success);

	printf("Flow = %lf\n",LARGE_NUM);

	for(y=SAD_SIZE; y<size.height-SAD_SIZE; y++) {
		temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[0];

		for(x=DISPARITY_MAX; x<size.width-SAD_SIZE; x++) {
			if(temp_fp[x]*CORRECTION > 255) {
				temp_fp[x] = 255;
			} else {
				temp_fp[x] *= CORRECTION;
			}
		}
	}

	return(fpImage);
}

int main(int argc, char *argv[]){
	char left[]  = "t_img01.jpg";
	char right[] = "t_img03.jpg";
	char dst[] = "dst.jpg";
	char window_left[] = "LEFT";
	char window_right[]= "RIGHT";
	char window_parallax[]= "PARALLAX";
	cvNamedWindow(window_left,CV_WINDOW_AUTOSIZE);
	cvNamedWindow(window_right,CV_WINDOW_AUTOSIZE);
	cvNamedWindow(window_parallax,CV_WINDOW_AUTOSIZE);

	IplImage *left_image = cvLoadImage(left,0);
	IplImage *right_image  = cvLoadImage(right,0);
	IplImage *parallax_image = alpha_cut(left_image,right_image);

	cvShowImage(window_left,left_image);
	cvShowImage(window_right,right_image);
	cvShowImage(window_parallax,parallax_image);

	cvSaveImage(dst,parallax_image);

	cvWaitKey(0);
	cvDestroyAllWindows();

	return 0;
}
